# Copyright (C) 2022 Keyu Tian. All rights reserved.
import threading
from functools import wraps
from multiprocessing import Pool as ProcPool
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool
from os import getpid

"""
from https://stackoverflow.com/questions/46045956/whats-the-difference-between-threadpool-vs-pool-in-the-multiprocessing-module
The rule of thumb is:
- IO bound jobs -> multiprocessing.pool.ThreadPool
- CPU bound jobs -> multiprocessing.Pool
- Hybrid jobs -> depends on the workload, I usually prefer the multiprocessing.Pool due to the advantage process isolation brings
"""
__all__ = ['to_parallelize', 'cpu_count', 'getpid', 'gettid', 'mt_proc', 'mt_thread']


def to_parallelize(segsize, ratio):
    cores = cpu_count()
    return cores >= 4 and segsize > ratio * cores
    

def mt_proc(chunksize=1):
    def decorator(func):
        @wraps(func)
        def wrapper(arg_list):
            with ProcPool(cpu_count()) as pool:
                rets = dict(pool.starmap(func, arg_list, chunksize=chunksize))
            return rets
        return wrapper
    
    return decorator


def mt_thread(chunksize=1):
    def decorator(func):
        @wraps(func)
        def wrapper(arg_list):
            with ThreadPool(cpu_count()) as pool:
                rets = list(pool.starmap(func, arg_list, chunksize=chunksize))
            return rets
        return wrapper
    
    return decorator


def gettid(naive=True):
    return threading.get_native_id() if naive else threading.get_ident()




#!/usr/bin/python3
# Copyright (C) 2022 Keyu Tian. All rights reserved.
import argparse
import glob
import os
import re
import sys
from functools import partial

import chardet

from toolky.core import to_parallelize, mt_thread, cpu_count

__all__ = ['main']


def find(f_name, regex, key_str):
    num_lines = 0
    try:
        with open(f_name, 'rb') as fp:
            byte_data = fp.read()
            lines = byte_data.decode(chardet.detect(byte_data)['encoding']).splitlines()
            contains = []
            for i, line in enumerate(lines):
                num_lines += 1
                if regex.match(line) is not None:
                    contains.append(i + 1)
        if len(contains) != 0:
            sys.stdout.write(f'`{key_str}\' found in  {f_name:24s}:  [line {str(contains).strip("[]")}]\n')
    except Exception as e:
        sys.stderr.write(f'[{f_name}: read err]\n')
        raise e
        pass

    return num_lines


@mt_thread(10)
def mp_find(file_name, regex, key_str):
    return find(file_name, regex, key_str)


def main():
    DIR_IGNORED = {'__pycache__', '__pypackages__', '.git', '.idea'}
    TARGET_POSTFIX = {
        'c', 'cc', 'cpp', 'cxx', 'cs', 'h', 'hpp', 'f', 'for',
        'py', 'java', 'kt', 'm',
        'js', 'ts', 'html', 'css', 'php',
        'yaml', 'sh',
        'json', 'txt', 'md',
    }
    
    gg = partial(glob.glob, recursive=True)
    ggi = partial(glob.iglob, recursive=True)
    
    desc = 'Search for some texts. [Example:  %(prog)s  "./dir/*.py"  "torch.*fft"  "3\\*3 == 9" ]'
    parser = argparse.ArgumentParser(prog='sea', description=desc)
    parser.add_argument('position', metavar='D/F', type=str, nargs='?', help='directory name or file name to be searched (wildcard * and ** is supported; default: cwd `.`)', default='./')
    parser.add_argument('keys', metavar='K', type=str, nargs='+', help='keys to search (wildcard * is supported)')
    parser.add_argument('-mp', '--mp', action='store_true', help='with multiprocessing')
    parser.add_argument('-nr', '--nr', action='store_true', help='not recursively')
    args = parser.parse_args()

    if args.position == '/':
        raise AttributeError('search from the root would be too costly!')
    
    dirs_and_files, keys = gg(os.path.expanduser(args.position)), args.keys

    files = set()
    for df in dirs_and_files:
        if os.path.isdir(df) and df not in DIR_IGNORED:
            f = filter(os.path.isfile, ggi(os.path.join(df, '**', '*')))
            files.update(set(f))
        elif os.path.isfile(df):
            files.add(df)

    files = sorted(filter(lambda fname: fname.split('.')[-1].lower() in TARGET_POSTFIX, files))
    key_str = ' | '.join(keys)
    
    t_keys = []
    s = chr(92)
    rand = 'uaweuyyawebubhdyugsaJYuaweiluidnuaeiu'
    for key in keys:
        key = key.replace(f'{s}*', rand)
        for ch in [s, '.', '^', '$', '*', '+', '?', '[', ']', '|', '{', '}', '(', ')']:
            key = key.replace(ch, s + ch)
        key = key
        t_keys.append(key)
    ks = [f'(.*{tk.replace("*", ".*").replace(rand, f"{s}*")})' for tk in t_keys]
    regex = re.compile('|'.join(ks))
    
    mp = args.mp or to_parallelize(len(files), 100)
    if mp:
        lines = mp_find([(f, regex, key_str) for f in files])
    else:
        lines = [find(f, regex, key_str) for f in files]
    
    print(f'\n[in {args.position}] #files: {len(files)}, #lines: {sum(lines)}' + (f' (w/ mt{cpu_count()})' if mp else ''))


if __name__ == '__main__':
    main()



# Copyright (C) 2022 Keyu Tian. All rights reserved.
import setuptools
from toolky import __version__


with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='toolky',
    version=__version__,
    author='Keyu Tian',
    author_email='tiankeyu.00@gmail.com',
    description='tools',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/keyu-codebase/toolky',
    packages=setuptools.find_packages(),
    package_data={
        # 引入任何包下面的 *.txt、*.rst 文件
        # "": ["*.txt", "*.rst"],
        # 引入 hello 包下面的 *.msg 文件
        # "hello": ["*.msg"],
    },
    platforms=['all'],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    python_requires='>=3.6',
    install_requires=[
        'chardet',
        # 'pathos',
    ],
    entry_points={
        'console_scripts': [
            'sea = toolky.cli.search:main'
        ]
    }
)





import functools
import os
from typing import List
from typing import Union

import torch
import torch.distributed as tdist
import torch.multiprocessing as mp

__rank, __local_rank, __world_size, __device = 0, 0, 1, 'cpu'
__initialized = False


def initialized():
    return __initialized


def initialize(backend='nccl'):
    # ref: https://github.com/open-mmlab/mmcv/blob/master/mmcv/runner/dist_utils.py#L29
    if mp.get_start_method(allow_none=True) is None:
        mp.set_start_method('spawn')
    global_rank, num_gpus = int(os.environ.get('RANK', 'error')), torch.cuda.device_count()
    local_rank = global_rank % num_gpus
    # assert local_rank == os.environ['LOCAL_RANK']   # torch>1.10
    torch.cuda.set_device(local_rank)
    tdist.init_process_group(backend=backend)
    
    global __rank, __local_rank, __world_size, __device, __initialized
    __local_rank = local_rank
    __rank, __world_size = tdist.get_rank(), tdist.get_world_size()
    __device = torch.empty(1).cuda().device
    __initialized = True
    
    assert tdist.is_initialized(), 'torch.distributed is not initialized!'


def get_rank():
    return __rank


def get_local_rank():
    return __local_rank


def get_world_size():
    return __world_size


def get_device():
    return __device


def is_master():
    return __rank == 0


def is_local_master():
    return __local_rank == 0


def parallelize(net, syncbn=False):
    if syncbn:
        net = torch.nn.SyncBatchNorm.convert_sync_batchnorm(net)
    net = net.cuda()
    net = torch.nn.parallel.DistributedDataParallel(net, device_ids=[get_local_rank()], find_unused_parameters=False, broadcast_buffers=False)
    return net


def new_group(ranks: List[int]):
    return tdist.new_group(ranks=ranks)


def barrier():
    tdist.barrier()


def allreduce(t: torch.Tensor) -> None:
    if not t.is_cuda:
        cu = t.detach().cuda()
        tdist.all_reduce(cu)
        t.copy_(cu.cpu())
    else:
        tdist.all_reduce(t)


def allgather(t: torch.Tensor, cat=True) -> Union[List[torch.Tensor], torch.Tensor]:
    if not t.is_cuda:
        t = t.cuda()
    ls = [torch.empty_like(t) for _ in range(__world_size)]
    tdist.all_gather(ls, t)
    if cat:
        ls = torch.cat(ls, dim=0)
    return ls


def broadcast(t: torch.Tensor, src_rank) -> None:
    if not t.is_cuda:
        cu = t.detach().cuda()
        tdist.broadcast(cu, src=src_rank)
        t.copy_(cu.cpu())
    else:
        tdist.broadcast(t, src=src_rank)


def dist_fmt_vals(val, fmt: Union[str, None] = '%.2f') -> Union[torch.Tensor, List]:
    ts = torch.zeros(__world_size)
    ts[__rank] = val
    allreduce(ts)
    if fmt is None:
        return ts
    return [fmt % v for v in ts.cpu().numpy().tolist()]


def master_only(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        force = kwargs.pop('force', False)
        if force or is_master():
            ret = func(*args, **kwargs)
        else:
            ret = None
        barrier()
        return ret
    return wrapper


def local_master_only(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        force = kwargs.pop('force', False)
        if force or is_local_master():
            ret = func(*args, **kwargs)
        else:
            ret = None
        barrier()
        return ret
    return wrapper
